/*
 * Copyright (c) 2015 ARM Limited. All rights reserved.
 * SPDX-License-Identifier: Apache-2.0
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include "mbed-client-mdk/m2mconnectionhandlerpimpl.h"
#include "mbed-client/m2mconnectionobserver.h"
#include "mbed-client/m2mconstants.h"
#include "mbed-client/m2msecurity.h"
#include "mbed-client/m2mconnectionhandler.h"

#include "minar/minar.h"

using mbed::util::FunctionPointer0;
using mbed::util::FunctionPointer2;

static M2MConnectionHandlerPimpl *Ptr_BSD[BSD_NUM_SOCKS] = { NULL };
static M2MConnectionHandlerPimpl *Prt_DNS;

static void net_dns_notify (netDNSc_Event event, const NET_ADDR *addr) {
    minar::Scheduler::postCallback(FunctionPointer2<void, netDNSc_Event, const NET_ADDR *>(Prt_DNS, &M2MConnectionHandlerPimpl::dns_handler).bind(event, addr));
}

M2MConnectionHandlerPimpl::M2MConnectionHandlerPimpl(M2MConnectionHandler* base, M2MConnectionObserver &observer,
                                                     M2MConnectionSecurity* sec,
                                                     M2MInterface::BindingMode mode,
                                                     M2MInterface::NetworkStack stack)
:_base(base),
 _observer(observer),
 _security_impl(sec),
 _use_secure_connection(false),
 _binding_mode(mode),
 _network_stack(stack),
 _resolved(true),
 _is_handshaking(false),
 _is_listening(false)
{
    _socket_address = (M2MConnectionObserver::SocketAddress *)malloc(sizeof(M2MConnectionObserver::SocketAddress));
    memset(_socket_address, 0, sizeof(M2MConnectionObserver::SocketAddress));
    _socket_address->_address = _received_address;

    int socket_family = AF_UNSPEC;
    unsigned long fnbio = 1;  // Non-blocking

    switch(_network_stack) {
        case M2MInterface::Uninitialized:
            break;
        case M2MInterface::LwIP_IPv4:
            break;
        case M2MInterface::LwIP_IPv6:
            break;
        case M2MInterface::Reserved:
            break;
        case M2MInterface::Nanostack_IPv6:
            break;
        case M2MInterface::MDK_Network_IPv4:
            socket_family = AF_INET;
            break;
        case M2MInterface::MDK_Network_IPv6:
            socket_family = AF_INET6;
            break;
        case M2MInterface::Unknown:
            break;
        default:
            break;
    }

    memset(_receive_buffer,0,sizeof(_receive_buffer));

    if(_binding_mode == M2MInterface::TCP ||
       _binding_mode == M2MInterface::TCP_QUEUE) {
        _socket = socket(socket_family, SOCK_STREAM, IPPROTO_TCP);
    } else {
        _socket = socket(socket_family, SOCK_DGRAM, IPPROTO_UDP);
    }
    if((_socket > 0) && (_socket <= BSD_NUM_SOCKS)) {
        ioctlsocket(_socket, FIONBIO, &fnbio);
        Ptr_BSD[_socket-1] = this;
    } else {
        closesocket(_socket);
        _socket = 0;
    }
}

M2MConnectionHandlerPimpl::~M2MConnectionHandlerPimpl()
{
    if((_socket > 0) && (_socket <= BSD_NUM_SOCKS)) {
        closesocket(_socket);
        Ptr_BSD[_socket-1] = NULL;
    }
    if(_socket_address) {
        free(_socket_address);
    }

    delete _security_impl;
}

bool M2MConnectionHandlerPimpl::bind_connection(const uint16_t listen_port)
{
    SOCKADDR_IN6 addr;
    int res = BSD_ERROR;

    memset(&addr, 0, sizeof(addr));
    if(_network_stack == M2MInterface::MDK_Network_IPv4) {
        ((SOCKADDR_IN *)&addr)->sin_family = AF_INET;
        ((SOCKADDR_IN *)&addr)->sin_port = htons(listen_port);
    } else if(_network_stack == M2MInterface::MDK_Network_IPv6) {
        addr.sin6_family = AF_INET6; 
        addr.sin6_port = htons(listen_port);
    }
    res = bind(_socket, (SOCKADDR *)&addr, sizeof(addr));

    return res == BSD_SUCCESS;
}

bool M2MConnectionHandlerPimpl::resolve_server_address(const String& server_address,
                                                      const uint16_t server_port,
                                                      M2MConnectionObserver::ServerType server_type,
                                                      const M2MSecurity* security)
{
    netStatus status = netError;
    int16_t addr_type;

    _security = security;

    if(_resolved) {
        _resolved = false;
        _server_address = server_address;
        _server_port = server_port;
        _server_type = server_type;

        if(_network_stack == M2MInterface::MDK_Network_IPv6) {
            addr_type = NET_ADDR_IP6;
        } else {
            addr_type = NET_ADDR_IP4;
        }
        Prt_DNS = this;
        status = netDNSc_GetHostByName(_server_address.c_str(), addr_type, net_dns_notify);
    }

    return status == netOK;
}

bool M2MConnectionHandlerPimpl::send_data(uint8_t *data,
                                     uint16_t data_len,
                                     sn_nsdl_addr_s *address)
{
    int res = BSD_ERROR;

    if(address == NULL || data == NULL) {
        return false;
    }
    if(_use_secure_connection) {
        if(_security_impl->send_message(data, data_len) > 0) {
            return true;
        } else {
            return false;
        }
    } else {
        int n;
        if(_binding_mode == M2MInterface::TCP ||
           _binding_mode == M2MInterface::TCP_QUEUE) {
            //We need to "shim" the length in front
            uint16_t d_len = data_len+4;
            uint8_t* d = (uint8_t*)malloc(data_len+4);

            d[0] = (data_len >> 24 )& 0xff;
            d[1] = (data_len >> 16 )& 0xff;
            d[2] = (data_len >> 8 )& 0xff;
            d[3] = data_len & 0xff;
            memmove(d+4, data, data_len);
            n = send(_socket, (const char*)d, d_len, 0);
            if (n == d_len) {
                res = BSD_SUCCESS;
            }
            free(d);
        } else {
          n = sendto(_socket, (const char*)data, data_len, 0, (const SOCKADDR*)&_resolved_Address, sizeof(_resolved_Address));
          if (n == data_len) {
              res = BSD_SUCCESS;
          }
        }
    }
    if (res != BSD_SUCCESS) {
        return false;
    }
    minar::Scheduler::postCallback(FunctionPointer0<void>(this, &M2MConnectionHandlerPimpl::send_handler).bind());
    return true;
}

void M2MConnectionHandlerPimpl::send_handler()
{
    _observer.data_sent();
}

bool M2MConnectionHandlerPimpl::start_listening_for_data()
{
    _is_listening = true;
    return true;
}

void M2MConnectionHandlerPimpl::stop_listening()
{
    _is_listening = false;
}

int M2MConnectionHandlerPimpl::send_to_socket(const unsigned char *buf, size_t len)
{
    int n;

    if(_binding_mode == M2MInterface::TCP ||
       _binding_mode == M2MInterface::TCP_QUEUE) {
        n = send(_socket, (const char*)buf, len, 0);
    } else {
        n = sendto(_socket, (const char*)buf, len, 0, (const SOCKADDR*)&_resolved_Address, sizeof(_resolved_Address));
    }

    if(n == BSD_ERROR_WOULDBLOCK) {
        return M2MConnectionHandler::CONNECTION_ERROR_WANTS_WRITE;
    } else if(n != len) {
        return -1;
    } else {
        minar::Scheduler::postCallback(FunctionPointer0<void>(this, &M2MConnectionHandlerPimpl::send_handler).bind());
        return len;
    }
}

int M2MConnectionHandlerPimpl::receive_from_socket(unsigned char *buf, size_t len)
{
    int n;

    if(_binding_mode == M2MInterface::TCP ||
       _binding_mode == M2MInterface::TCP_QUEUE) {
        n = recv(_socket, (char *)buf, len, 0);
    } else {
        n = recvfrom(_socket, (char*)buf, len, 0, NULL, 0);
    }

    if(n == BSD_ERROR_WOULDBLOCK) {
        return M2MConnectionHandler::CONNECTION_ERROR_WANTS_READ;
    } else if(n < 0) {
        return -1;
    } else {
        return n;
    }
}

void M2MConnectionHandlerPimpl::receive_handshake_handler()
{
    memset(_receive_buffer, 0, BUFFER_LENGTH);
    if(_is_handshaking) {
        int ret = _security_impl->continue_connecting();
        if(ret == M2MConnectionHandler::CONNECTION_ERROR_WANTS_READ) { //We wait for next readable event
            return;
        } else if(ret == 0) {
            _is_handshaking = false;
            _use_secure_connection = true;
            _observer.address_ready(*_socket_address,
                                    _server_type,
                                    _server_port);
        } else if(ret < 0) {
            //TODO: Socket error in SSL handshake,
            // Define error code.
            _is_handshaking = false;
            _observer.socket_error(4);
        }
    }
}

void M2MConnectionHandlerPimpl::receive_handler()
{
    memset(_receive_buffer, 0, BUFFER_LENGTH);
    size_t receive_length = sizeof(_receive_buffer);

    if(_use_secure_connection) {
        int rcv_size = _security_impl->read(_receive_buffer, receive_length);
        if(rcv_size >= 0) {
            _observer.data_available((uint8_t*)_receive_buffer,
                                     rcv_size, *_socket_address);
        } else if (M2MConnectionHandler::CONNECTION_ERROR_WANTS_READ != rcv_size) {
            _observer.socket_error(1);
            return;
        }
    } else {
        int n;
        if(_binding_mode == M2MInterface::TCP ||
           _binding_mode == M2MInterface::TCP_QUEUE) {
            n = recv(_socket, (char *)_receive_buffer, receive_length, 0);
        } else {
            n = recvfrom(_socket, (char*)_receive_buffer, receive_length, 0, NULL, 0);
        }
        if (n >= 0) {
            receive_length = n;
            memset(_socket_address,0,sizeof(M2MConnectionObserver::SocketAddress));
            //TODO: Current support only for IPv4, add IPv6 support
            if(_network_stack == M2MInterface::MDK_Network_IPv4) {
                _socket_address->_address = &((SOCKADDR_IN *)&_resolved_Address)->sin_addr;
                _socket_address->_length = 4;
            } else if(_network_stack == M2MInterface::MDK_Network_IPv6) {
                _socket_address->_address = &_resolved_Address.sin6_addr;
                _socket_address->_length = 16;
            }
            _socket_address->_port = _socket_address->_port;
            _socket_address->_stack = _network_stack;
            // Send data for processing.
            if(_binding_mode == M2MInterface::TCP ||
               _binding_mode == M2MInterface::TCP_QUEUE) {
                //We need to "shim" out the length from the front
                if(receive_length > 4) {
                    uint64_t len = (_receive_buffer[0] << 24 & 0xFF000000) + (_receive_buffer[1] << 16 & 0xFF0000);
                    len += (_receive_buffer[2] << 8 & 0xFF00) + (_receive_buffer[3] & 0xFF);
                    if(len > 0) {
                        uint8_t* buf = (uint8_t*)malloc(len);
                        if(buf) {
                            memmove(buf, _receive_buffer+4, len);
                            // Observer for TCP plain mode
                            _observer.data_available(buf,len,*_socket_address);
                            free(buf);
                        }
                    }
                } else {
                    _observer.socket_error(1);
                }
            } else { // Observer for UDP plain mode
                _observer.data_available((uint8_t*)_receive_buffer,
                                         receive_length, *_socket_address);
            }
        } else {
            // Socket error in receiving
            _observer.socket_error(1);
        }
    }
}

void M2MConnectionHandlerPimpl::dns_handler(netDNSc_Event event, const NET_ADDR *addr)
{
    _resolved = true;

    if(event != netDNSc_EventSuccess) {
        _observer.socket_error(2);
        return;
    }

    memset(&_resolved_Address, 0, sizeof(_resolved_Address));
    if(_network_stack == M2MInterface::MDK_Network_IPv4) {
        ((SOCKADDR_IN *)&_resolved_Address)->sin_family = AF_INET;
        ((SOCKADDR_IN *)&_resolved_Address)->sin_port = htons(_server_port);
        memcpy(&((SOCKADDR_IN *)&_resolved_Address)->sin_addr, addr->addr, NET_ADDR_IP4_LEN);
    } else if(_network_stack == M2MInterface::MDK_Network_IPv6) {
        _resolved_Address.sin6_family = AF_INET6;
        _resolved_Address.sin6_port = htons(_server_port);
        memcpy(&_resolved_Address.sin6_addr, addr->addr, NET_ADDR_IP6_LEN);
    }

    memset(_socket_address,0,sizeof(M2MConnectionObserver::SocketAddress));
    if(_network_stack == M2MInterface::MDK_Network_IPv4) {
        _socket_address->_address = &((SOCKADDR_IN *)&_resolved_Address)->sin_addr;
        _socket_address->_length = 4;
    } else if(_network_stack == M2MInterface::MDK_Network_IPv6) {
        _socket_address->_address = &_resolved_Address.sin6_addr;
        _socket_address->_length = 16;
    } else {
        _observer.socket_error(2);
        return;
    }
    _socket_address->_stack = _network_stack;
    _socket_address->_port = _server_port;

    if(_binding_mode == M2MInterface::TCP ||
       _binding_mode == M2MInterface::TCP_QUEUE){
        connect(_socket, (const SOCKADDR*)&_resolved_Address, sizeof(_resolved_Address));
    }

    if(_security) {
        if(_security->resource_value_int(M2MSecurity::SecurityMode) == M2MSecurity::Certificate ||
           _security->resource_value_int(M2MSecurity::SecurityMode) == M2MSecurity::Psk) {
            if(_security_impl != NULL) {
                _security_impl->reset();
                _security_impl->init(_security);
                _is_handshaking = true;
                if(_security_impl->start_connecting_non_blocking(_base) < 0) {
                    //TODO: Socket error in SSL handshake,
                    // Define error code.
                    _is_handshaking = false;
                    _observer.socket_error(4);
                    return;
                }
            }
        }
    }
    if(!_is_handshaking) {
        _observer.address_ready(*_socket_address,
                                _server_type,
                                _server_port);
    }
}

void M2MConnectionHandlerPimpl::handle_connection_error(int /*error*/)
{
    //This will come from M2MConnectionSecurity class
    _observer.socket_error(4);
}

extern "C" void net_bsd_notify (int sock, uint8_t event) {
    M2MConnectionHandlerPimpl::net_bsd_notify(sock, event);
}

void M2MConnectionHandlerPimpl::net_bsd_notify (int sock, uint8_t event) {
    M2MConnectionHandlerPimpl *ptr;

    if((sock > 0) && (sock <= BSD_NUM_SOCKS)) {
        ptr = Ptr_BSD[sock-1];
        if(ptr) {
          if(event & BSD_EVT_DATA) {
              if(ptr->_is_handshaking) {
                  minar::Scheduler::postCallback(FunctionPointer0<void>(ptr, &M2MConnectionHandlerPimpl::receive_handshake_handler).bind());
              } else if(ptr->_is_listening) {
                  minar::Scheduler::postCallback(FunctionPointer0<void>(ptr, &M2MConnectionHandlerPimpl::receive_handler).bind());
              }
          }
        }
    }
}
